
from functools import reduce
from django.db.models import Q
from .serializer import Serializer
from .util import parse_date, parse_datetime

int_op_map = {'equals': 'exact',
              'contains': 'exact',
              'startswith': 'gte',
              'endswith': 'lte'}

str_op_map = {'equals': 'exact',
              'gt': 'startswith',
              'gte': 'startswith',
              'lt': 'endswith',
              'lte': 'endswith'}

date_op_map = {'startswith': 'gte',
               'endswith': 'lte'}

class DjangoSerializer(Serializer):
    def __init__(self, model):
        Serializer.__init__(self)
        self.model = model

    def logical_and(self, terms):
        terms = [t for t in terms if t]
        return reduce(lambda x, y: x.__and__(y), terms, Q())

    def logical_or(self, terms):
        terms = [t for t in terms if t]
        if not terms:
            return Q()
        return reduce(lambda x, y: x.__or__(y), terms)

    def logical_not(self, terms):
        if len(terms) == 1:
            return ~terms[0]
        return ~self.logical_and(terms)

    def boolean_term(self, selector, operator, data):
        if data.lower() not in ['true', 'false']:
            return ~Q()
        value = data.lower() == 'true'
        return Q(**{selector: value})

    def int_term(self, selector, operator, data):
        try:
            value = int(data)
        except ValueError:
            return Q()
        operator = int_op_map.get(operator, operator)
        if operator == 'exact':
            return Q(**{selector: value})
        return Q(**{selector+'__'+operator: value})

    def str_term(self, selector, operator, data):
        operator = str_op_map.get(operator, operator)
        return Q(**{selector+'__'+operator: data})

    def lcstr_term(self, selector, operator, data):
        operator = str_op_map.get(operator, operator)
        return Q(**{selector+'__i'+operator: data})

    def date_datetime_common(self, selector, operator, thedatetime):
        if not thedatetime:
            return Q()
        operator = date_op_map.get(operator, operator)
        if operator in ('contains', 'equals'):
            return Q(**{selector+'__year': thedatetime.year,
                        selector+'__month': thedatetime.month,
                        selector+'__day': thedatetime.day})
        return Q(**{selector+'__'+operator: thedatetime})

    def date_term(self, selector, operator, data):
        thedate = parse_date(data)
        return self.date_datetime_common(selector, operator, thedate)

    def datetime_term(self, selector, operator, data):
        thedatetime = parse_datetime(data)
        result = self.date_datetime_common(selector, operator, thedatetime)
        if operator != 'equals' or not result:
            return result
        return result&Q(**{selector+'__hour': thedatetime.hour,
                           selector+'__minute': thedatetime.minute})

    def term(self, name, operator, data):
        if operator == 'any':
            return Q()

        cls, alias = self.model.get_class_from_fullname(name)
        handler = cls.get_field_handler_from_alias(alias)
        selector = self.model.get_selector_from_fullname(name)

        type_map = {'BOOL': self.boolean_term,
                    'INT': self.int_term,
                    'STR': self.str_term,
                    'LCSTR': self.lcstr_term,
                    'DATE': self.date_term,
                    'DATETIME': self.datetime_term}

        func = type_map.get(handler.db_type)
        if not func:
            raise TypeError('unsupported field type: '+repr(field_type))
        return func(selector, operator, handler.prepare(data))



class Parser(object):
    """
    The base class for all parsers.
    """

    def __init__(self, token_list):
        self.token_list = token_list
        self._reset()

    def _reset(self):
        self.offset = 0
        self.line = 0
        self.error = ''

    def _get_next_token(self):
        if len(self.input) <= self.offset:
            return 'EOF', None

        # Walk through the list of tokens, trying to find a match.
        for token_name, token_regex in self.token_list:
            match = token_regex.match(self.input, self.offset)
            if not match:
                continue

            string = match.group(0)
            self.offset += len(string)
            self.line += string.count('\n')
            return token_name, match

        # Ending up here no matching token was found.
        return None, None



import re
from ..refs import get_subclasses
from django_find import models
from collections import OrderedDict
from .parser import Parser
from ..dom import Group, And, Or, Not, Term

operators = OrderedDict((
    ('!=', 'notequals'),
    ('<>', 'notequals'),
    ('>=', 'gte'),
    ('=>', 'gte'),
    ('>', 'gt'),
    ('<=', 'lte'),
    ('=<', 'lte'),
    ('<', 'lt'),
    (':', 'contains'),
    ('!:', 'notcontains'),
    ('=', 'equals'),
))

operators_str = '|'.join(list(operators.keys()))
tokens = [
    ('and', re.compile(r'and\b', re.I)),
    ('or', re.compile(r'or\b', re.I)),
    ('not', re.compile(r'not\b', re.I)),
    ('openbracket', re.compile(r'\(')),
    ('closebracket', re.compile(r'\)')),
    ('whitespace', re.compile(r'\s+')),
    ('field', re.compile(r'([\w\-]+)({})'.format(operators_str))),
    ('word', re.compile(r'"([^"]*)"')),
    ('word', re.compile(r'([^"\s\\\'\)]+)')),
    ('unknown', re.compile(r'.'))
]

def open_scope(scopes, scope):
    scopes.append(scopes[-1].add(scope))

def close_scope(scopes):
    while scopes[-1].auto_leave_scope() and not scopes[-1].is_root:
        scopes.pop()

def op_from_word(word):
    if word.startswith('^') and word.endswith('$'):
        return word[1:-1], 'equals'
    elif word.startswith('^'):
        return word[1:], 'startswith'
    elif word.endswith('$'):
        return word[:-1], 'endswith'
    return word, 'contains'

def get_term_from_op(field, operator, value):
    op = operators.get(operator)

    if op == 'contains':
        value, op = op_from_word(value)
    if op == 'notcontains':
        value, op = op_from_word(value)
        op = 'not'+op

    if op.startswith('not'):
        return Not(Term(field, op[3:], value))
    return Term(field, op, value)

class QueryParser(Parser):
    def __init__(self, fields, default):
        """
        Fields is a map that translates aliases to something like
        Book.author.
        """
        Parser.__init__(self, tokens)
        self.fields = fields
        self.default = default or fields
        for name in self.default:
            if name not in self.fields:
                raise AttributeError('constructor argument "default" contains'\
                    +' "{}", which is not also in "fields"'.format(name))

    def parse_word(self, scopes, match):
        self.parse_or(scopes, ())
        child = Or()
        for name in self.default:
            name = self.fields[name]
            value, operator = op_from_word(match.group(1))
            child.add(Term(name, operator, value))
        scopes[-1].add(child)
        close_scope(scopes)

    def parse_field(self, scopes, match):
        field_name = match.group(1).lower()
        field = self.fields.get(field_name)
        if field is None:
            self.parse_word(scopes, match)
            return

        # A field value is required.
        op = match.group(2)
        token, match = self._get_next_token()
        try:
            value = match.group(1)
        except IndexError:
            return

        for subclass in get_subclasses(models.Searchable):
            try:
                for k, v in subclass._meta.get_field(field_name).choices:
                    if value == v:
                        value = k
                        break
            except:
                continue

        term = get_term_from_op(field, op, value)
        scopes[-1].add(term)
        close_scope(scopes)

    def parse_boolean(self, scopes, dom_cls, match):
        try:
            if scopes[-1].is_logical() and dom_cls.precedence() < scopes[-1].precedence():
                scopes.pop()
            if type(scopes[-1]).__name__ == dom_cls.__name__:
                return
            last_term = scopes[-1].pop()
        except IndexError:
            open_scope(scopes, dom_cls())
        else:
            open_scope(scopes, dom_cls(last_term))

    def parse_and(self, scopes, match):
        self.parse_boolean(scopes, And, match)

    def parse_or(self, scopes, match):
        self.parse_boolean(scopes, Or, match)

    def parse_term(self, token, scopes, match):
        try:
            parse_func = getattr(self, 'parse_'+token)
            parse_func(scopes, match)
        except AttributeError:
            pass

    def add_logical_scope(self, scopes, match):
        """
        Parses a logical scope within a query.
        A logical scope is required to handle 'not' statements that are at the root scope level.
        This method processes tokens within the scope, handling open and close brackets,
        whitespace, and words appropriately.
        Args:
            scopes (list): The list of current scopes being processed.
            match (re.Match): The current match object from the regular expression.
        Returns:
            None
        """
        self.parse_openbracket(scopes, match)
        stop_on_whitespace = False
        while True:
            token, match = self._get_next_token()
            if token == 'EOF':
                break
            elif token == 'whitespace':
                if stop_on_whitespace:
                    break
                else:
                    continue
            elif token == 'word':
                stop_on_whitespace = True
            self.parse_term(token, scopes, match)
        self.parse_closebracket(scopes, match)

    def parse_not(self, scopes, match):
        open_scope(scopes, Not())
        self.add_logical_scope(scopes, match)

    def parse_openbracket(self, scopes, match):
        open_scope(scopes, Group())

    def parse_closebracket(self, scopes, match):
        # Leave the current group.
        while type(scopes[-1]).__name__ != Group.__name__ and not scopes[-1].is_root:
            scopes.pop()
        if not scopes[-1].is_root:
            scopes.pop()

        # Auto-leave the parent scope (if necessary).
        close_scope(scopes)

    def parse(self, query):
        self._reset()
        self.input = query.strip()
        result = Group(is_root=True)
        scopes = [result]
        token, match = self._get_next_token()

        while token != 'EOF':
            self.parse_term(token, scopes, match)
            token, match = self._get_next_token()

        return result.optimize()




class Serializer(object):
    """
    Base class for all serializers.
    """

    def logical_root_group(self, root_group, terms):
        return self.logical_group(terms)

    def logical_group(self, terms):
        return self.logical_and(terms)



from builtins import str
from .tree import Node

operators = [
    'contains',
    'equals',
    'startswith',
    'endswith',
    'regex',
    'gt',
    'gte',
    'lt',
    'lte',
    'any'
]

class Group(Node):
    def translate_term_names(self, name_map):
        def translate(dom_obj):
            dom_obj.name = name_map.get(dom_obj.name, dom_obj.name)
        self.each(translate, Term)

    def get_term_names(self):
        """
        Returns a flat list of the names of all Terms in the query, in
        the order in which they appear. Filters duplicates.
        """
        field_names = []
        def collect_field_names(dom_obj):
            if not dom_obj.name in field_names:
                field_names.append(dom_obj.name)
        self.each(collect_field_names, Term)
        return field_names

    def auto_leave_scope(self):
        return False

    def optimize(self):
        children = [c.optimize() for c in self.children]
        self.children = [c for c in children if c is not None]
        children = []
        for child in self.children:
            if type(child) == type(self):
                for grandchild in child.children:
                    children.append(grandchild)
            else:
                children.append(child)
        self.children = children
        if not self.children and not self.is_root:
            return None
        if len(self.children) == 1 and not self.is_root:
            return self.children[0]
        return self

    def serialize(self, strategy):
        results = [c.serialize(strategy) for c in self.children]
        if self.is_root:
            return strategy.logical_root_group(self, results)
        return strategy.logical_group(results)

class And(Group):
    @classmethod
    def is_logical(self):
        return True

    @classmethod
    def precedence(self):
        return 2

    def serialize(self, strategy):
        return strategy.logical_and(c.serialize(strategy)
                                    for c in self.children)

class Or(Group):
    @classmethod
    def is_logical(self):
        return True

    @classmethod
    def precedence(self):
        return 1

    def serialize(self, strategy):
        return strategy.logical_or(c.serialize(strategy)
                                   for c in self.children)

class Not(Group):
    @classmethod
    def precedence(self):
        return 3

    def auto_leave_scope(self):
        return True

    def optimize(self):
        children = [c.optimize() for c in self.children]
        self.children = [c for c in children if c is not None]
        if not self.children and not self.is_root:
            return None
        return self

    def serialize(self, strategy):
        children = [c.serialize(strategy) for c in self.children]
        return strategy.logical_not(children)

class Term(Node):
    def __init__(self, name, operator, data):
        assert operator in operators, "unsupported operator {}".format(operator)
        Node.__init__(self)
        self.name = name
        self.operator = str(operator)
        self.data = str(data)

    def optimize(self):
        return self

    def each(self, func, node_type):
        if node_type is None or isinstance(self, node_type):
            func(self)

    def dump(self, indent=0):
        return [(indent * '  ')
                + self.__class__.__name__ + ': '
                + self.name + ' ' + self.operator + ' ' + repr(self.data)]

    def serialize(self, strategy):
        return strategy.term(self.name, self.operator, self.data)




from django import template
from django.template.loader import render_to_string

class SearchNode(template.Node):
    def __init__(self, queryset_var, fields):
        self.queryset_var = template.Variable(queryset_var)
        self.fields = fields

    def render(self, context):
        request = context['request']
        getvars = request.GET.copy()

        if 'q' in getvars:
            # Search, and store the resulting queryset in the current
            # context.
            query = getvars['q']
            queryset = self.queryset_var.resolve(context)
            q_obj = queryset.model.q_from_query(query, self.fields)
            context[self.queryset_var.var] = queryset.filter(q_obj)

        return render_to_string('django_find/form.html',
                                {'getvars': getvars})

def find(parser, token):
    contents = token.split_contents()
    if len(contents) < 2:
        raise template.TemplateSyntaxError(
            "%r tag requires at least 1 argument, " +
            "in the form of {%% %r model.objects.all [alias1 alias2 ...] %%}" % contents[0])

    return SearchNode(contents[1], contents[2:])

register = template.Library()
register.tag('find', find)





class Node(object):
    def __init__(self, children=None, is_root=False):
        if isinstance(children, Node):
            children = [children]
        self.is_root = is_root
        self.children = list(children) if children else []

    @classmethod
    def precedence(self):
        return 0

    @classmethod
    def is_logical(self):
        return False

    def add(self, child):
        self.children.append(child)
        return child

    def pop(self):
        return self.children.pop()

    def dump(self, indent=0):
        isroot = '(root)' if self.is_root else ''
        result = [(indent * '  ') + self.__class__.__name__ + isroot]
        for child in self.children:
            result += child.dump(indent+1)
        if self.is_root:
            return '\n'.join(result)
        return result

    def each(self, func, node_type=None):
        """
        Runs func once for every node in the object tree.
        If node_type is not None, only call func for nodes with the given
        type.
        """
        if node_type is None or isinstance(self, node_type):
            func(self)
        for child in self.children:
            child.each(func, node_type)



from .serializers.sql import SQLSerializer

def sql_from_dom(cls, dom, mode='SELECT', fullnames=None, extra_model=None):
    if not fullnames:
        fullnames = dom.get_term_names()
    if not fullnames:
        return 'SELECT * FROM (SELECT NULL) tbl WHERE 0', [], [] # Empty set
    primary_cls = cls.get_primary_class_from_fullnames(fullnames)
    serializer = SQLSerializer(primary_cls,
                               mode=mode,
                               fullnames=fullnames,
                               extra_model=extra_model)
    sql, args = dom.serialize(serializer)
    return sql, args, fullnames



from dateparser import parse

DATEPARSER_SETTING = {
    'TIMEZONE': 'UTC',
    'STRICT_PARSING': True
}

def parse_date(thedate):
    thedatetime = parse(thedate)
    if not thedatetime:
        return None
    return thedatetime.date()

def parse_datetime(thedate):
    return parse(thedate)



from django.db import connection

SQL_MAXINT=9223372036854775807 # SQLite maxint

def assert_positive_slice(slc):
    if (slc.start is not None and slc.start < 0) or \
       (slc.stop is not None and slc.stop < 0):
       raise IndexError("Negative indexing is not supported")

class PaginatedRawQuerySet(object):
    def __init__(self, model, raw_query, args=None, limit=None, offset=None):
        self.model = model
        self.raw_query = raw_query
        self.args = args if args else []
        self.limit = limit
        self.offset = offset or 0
        self.result_cache = None
        self.count_cache = None

    def __copy__(self):
        return self.__class__(self.model,
                              self.raw_query,
                              self.args,
                              limit=self.limit,
                              offset=self.offset)

    def _getslice(self, slc):
        assert_positive_slice(slc)
        qs = self.__copy__()
        qs.offset = slc.start or 0
        qs.limit = None if slc.stop is None else (slc.stop-qs.offset)
        return qs

    def _getindex(self, idx):
        if idx < 0:
            raise IndexError("Negative indexing is not supported")
        qs = self.__copy__()
        qs.offset = self.offset+idx if self.offset else idx
        qs.limit = 1
        return list(qs)[idx]

    def __getitem__(self, k):
        """
        Retrieves an item or slice from the set of results.
        """
        if isinstance(k, slice):
            return self._getslice(k)
        if isinstance(k, int):
            return self._getindex(k)
        raise TypeError

    @property
    def query(self):
        query = self.raw_query
        if self.limit is None:
            query += ' LIMIT '+str(SQL_MAXINT-self.offset)+' OFFSET '+str(int(self.offset))
        else:
            query += ' LIMIT '+str(int(self.limit))+' OFFSET '+str(int(self.offset))
        return query

    def __iter__(self):
        if self.result_cache is None:
            with connection.cursor() as cursor:
                cursor.execute(self.query, self.args)
                self.result_cache = cursor.fetchall()
        return iter(self.result_cache)

    def __len__(self):
        if self.count_cache is not None:
            return self.count_cache
        query = 'SELECT COUNT(*) FROM (' + self.query + ') c'
        with connection.cursor() as cursor:
            cursor.execute(query, self.args)
            row = cursor.fetchone()
            self.count_cache = int(row[0])
        return self.count_cache

    count = property(__len__) # For better compatibility to Django's QuerySet




import json
from collections import OrderedDict
from ..dom import Group, And, Or, Not, Term

class JSONParser(object):
    """
    Transforms a JSON string into a DOM. The DOM is identical to what
    QueryParser generates. Example JSON input::

        {
            "Device":
            {
                "Hostname":
                    [
                        [["contains": "s-"],["contains": "-ea1"]],
                        [["startswith", ""]]
                    ],
                "Tags":
                    [
                        [["neq":"asdasd"]]
                    ]
            }
            "Component":
            {
                "Slot": [[]]
            }
        }
    """

    def parse_operators(self, termgroup, term, fieldname):
        for operator, value in term:
            if operator.startswith('not'):
                term = Not(Term(fieldname, operator[3:], value))
            else:
                term = Term(fieldname, operator, value)
            termgroup.add(term)

    def parse_terms(self, fieldgroup, terms, fieldname):
        for term in terms:
            termgroup = And()
            fieldgroup.add(termgroup)
            if not term:
                termgroup.add(Term(fieldname, 'any', ''))
                continue
            self.parse_operators(termgroup, term, fieldname)

    def parse_criteria(self, clsgroup, criteria, clsname):
        for fieldname, terms in criteria.items():
            fieldname = clsname + '.' + fieldname
            fieldgroup = Or()
            clsgroup.add(fieldgroup)
            self.parse_terms(fieldgroup, terms, fieldname)

    def parse(self, json_string):
        json_tree = json.loads(json_string, object_pairs_hook=OrderedDict)
        result = Group(is_root=True)

        for clsname, criteria in json_tree.items():
            clsgroup = And()
            result.add(clsgroup)
            self.parse_criteria(clsgroup, criteria, clsname)

        return result.optimize()



"""
This module contains the Searchable mixin, the main public API of
django-find.
"""
from collections import OrderedDict
from django.apps import AppConfig
from django.db import models
from .parsers.query import QueryParser
from .parsers.json import JSONParser
from .serializers.django import DjangoSerializer
from .refs import get_subclasses, get_object_vector_to, get_object_vector_for
from .rawquery import PaginatedRawQuerySet
from .model_helpers import sql_from_dom
from .handlers import type_registry

class Searchable(object):
    """
    This class is a mixin for Django models that provides methods for
    searching the model using query strings and other tools.
    """

    searchable_labels = {} # Override the verbose_name for the given aliases
    searchable = () # Contains two-tuples, mapping aliases to Django selectors

    @classmethod
    def get_default_searchable(cls):
        return OrderedDict((f.name, f.name) for f in cls._meta.get_fields()
                           if not f.auto_created)

    @classmethod
    def get_searchable(cls):
        result = cls.get_default_searchable()
        if hasattr(cls, 'searchable'):
            result.update(OrderedDict(cls.searchable))
        return tuple(i for i in result.items() if i[1])

    @classmethod
    def get_caption_from_selector(cls, selector):
        caption = cls.searchable_labels.get(selector)
        if caption:
            return caption
        field = cls.get_field_from_selector(selector)[1]
        if hasattr(field, 'verbose_name'):
            return field.verbose_name
        return field.name.capitalize()

    @classmethod
    def get_field_handler_from_field(cls, field):
        if isinstance(field, models.ForeignKey):
            field = field.target_field
        for handler in type_registry:
            if handler.handles(cls, field):
                return handler
        msg = 'field {}.{} is of type {}'.format(cls.get_classname(),
                                                 field.name,
                                                 type(field))
        raise TypeError(msg + ', which has no field handler. Consider adding a '
                      + 'django_find.handlers.FieldHandler to the '
                      + 'django_find.handlers.type_registry. See the docs for'
                      + 'more information')

    @classmethod
    def get_aliases(cls):
        """
        Returns a list of the aliases, that is, the names of the
        fields that can be used in a query.
        """
        return list(OrderedDict(cls.get_searchable()).keys())
    
    @classmethod
    def get_classname(cls):
        """
        Returns the class name used to build fullnames and Django selectors.

        When the short class name is unique among all Searchable
        subclasses the bare name is returned (e.g. ``"Book"``).
        When another Searchable model has the same ``__name__``
        (different app), the Django app label is prepended to
        disambiguate (e.g. ``"myapp.Author"``), matching the format
        accepted by :meth:`get_class_from_fullname`.
        """
        short = cls.__name__
        # Check if any *other* Searchable subclass shares our __name__.
        for sub in get_subclasses(Searchable):
            if sub is cls or sub.__module__ == '__fake__':
                continue
            if sub.__name__ == short:
                return cls._meta.label   # e.g. "search_tests.Author"
        return short

    @classmethod
    def get_fullnames(cls, unique=False):
        """
        Like get_aliases(), but returns the aliases prefixed by the class
        name.
        """
        if unique:
            selectors = set()
            result = []
            for item in cls.get_searchable():
                selector = item[1]
                if selector in selectors:
                    continue
                selectors.add(selector)
                result.append(cls.get_classname()+'.'+item[0])
            return result
        else:
            aliases = cls.get_aliases()
            return [cls.get_classname()+'.'+alias for alias in aliases]

    @classmethod
    def table_headers(cls):
        selectors = set()
        result = []
        for item in cls.get_searchable():
            selector = item[1]
            if selector in selectors:
                continue
            selectors.add(selector)
            result.append(cls.get_caption_from_selector(selector))
        return result

    @classmethod
    def get_field_from_selector(cls, selector):
        """
        Given a django selector, e.g. device__metadata__name, this returns the class
        and the Django field of the model, as returned by Model._meta.get_field().
        Example::

            device__metadata__name -> (SeedDevice, SeeDevice.name)
        """
        if not '__' in selector:
            return cls, cls._meta.get_field(selector)

        model = cls
        while '__' in selector:
            model_name, selector = selector.split('__', 1)
            model = model._meta.get_field(model_name).remote_field.model

        return model, model._meta.get_field(selector)

    @classmethod
    def get_field_handler_from_alias(cls, alias):
        """
        Given an alias, e.g. 'host', 'name',
        this function returns the handler.FieldHandler.

        @type name: str
        @param name: e.g. 'address', or 'name'
        """
        selector = cls.get_selector_from_alias(alias)
        field = cls.get_field_from_selector(selector)[1]
        return cls.get_field_handler_from_field(field)

    @classmethod
    def get_field_handler_from_fullname(cls, fullname):
        """
        Given a fullname, e.g. 'Device.host', 'Author.name',
        this function returns the handler.FieldHandler.

        @type name: str
        @param name: e.g. 'address', or 'name'
        """
        thecls, alias = cls.get_class_from_fullname(fullname)
        return thecls.get_field_handler_from_alias(alias)

    @classmethod
    def get_selector_from_alias(cls, alias):
        """
        Given alias (not a fullname), this function returns the
        selector in the following form::

            component__device__host

        @type name: str
        @param name: e.g. 'address', or 'name'
        """
        return dict(cls.get_searchable())[alias]

    @classmethod
    def get_object_vector_to(cls, search_cls):
        return get_object_vector_to(cls, search_cls, Searchable)

    @classmethod
    def get_object_vector_for(cls, search_cls_list):
        return get_object_vector_for(cls, search_cls_list, Searchable)

    @classmethod
    def get_class_from_fullname(cls, fullname):
        """
        Given a name in the format "Model.field" or "app_label.Model.field",
        this function returns a tuple ``(ModelClass, alias)``.

        The short form ``Author.name`` works when only one Searchable
        subclass is named *Author*.  When two apps define a model with
        the same class name, use the qualified form
        ``app_label.Author.name``  (the ``app_label`` comes from
        ``Model._meta.label``, e.g. ``myapp.Author``).

        When ``cls`` (the model the method is called on) can narrow down
        the ambiguity — because exactly one match shares the same
        ``app_label`` — that match is preferred automatically.  This
        lets serializers resolve short names produced by the JSON parser
        without requiring every caller to pass qualified names.
        """
        if '.' not in fullname:
            raise AttributeError(
                'class name is required, format should be '
                '"Class.alias" or "app_label.Class.alias"'
            )

        # Split into class identifier and field alias.
        names = fullname.split('.')
        clsname = ".".join(names[:-1])  # e.g. "Author" or "myapp.Author"
        alias = names[-1]

        # Collect all Searchable subclasses, skipping Django internals.
        candidates = [
            sub for sub in get_subclasses(Searchable)
            if sub.__module__ != '__fake__'
        ]

        # First, try an exact match against the Django app label
        # (e.g. "myapp.Author" == Model._meta.label).  This is the
        # unambiguous, qualified form.
        for sub in candidates:
            if sub._meta.label == clsname:
                return sub, alias

        # Fall back to matching the short class name.
        matches = [sub for sub in candidates if sub.__name__ == clsname]
        if len(matches) == 1:
            return matches[0], alias
        if len(matches) > 1:
            # Try to disambiguate using the calling class's app_label.
            # When a serializer calls  cls.get_class_from_fullname(name),
            # cls is the "root" model of the query — models in the same
            # app are almost certainly the intended targets.
            same_app = [
                m for m in matches
                if m._meta.app_label == cls._meta.app_label
            ]
            if len(same_app) == 1:
                return same_app[0], alias

            labels = ', '.join(sorted(m._meta.label for m in matches))
            raise KeyError(
                f'Ambiguous class name "{clsname}" matches multiple '
                f'models: {labels}. Use the qualified form '
                f'"app_label.{clsname}.{alias}" to disambiguate.'
            )

        raise KeyError('no such class: ', clsname)

    @classmethod
    def get_selector_from_fullname(cls, fullname):
        """
        Given a name in the form 'Unit.hostname', this function returns
        a Django selector that can be used for filtering.
        Example (assuming the models are Book and Author)::

            Book.get_selector_from_fullname('Author.birthdate')
            # returns 'author__birthdate'

        Example for the models Blog, Entry, Comment::

            Blog.get_selector_from_fullname('Comment.author')
            # returns 'entry__comment__author'

        @type name: str
        @param name: The field to select for
        @rtype: str
        @return: The Django selector
        """
        # Get the target class and attribute by parsing the name.
        target_cls, alias = cls.get_class_from_fullname(fullname)
        selector = target_cls.get_selector_from_alias(alias)
        if target_cls == cls:
            return selector

        # Prefix the target by the class names.
        path_list = get_object_vector_to(cls, target_cls, Searchable)
        path = path_list[0]
        prefix = ''
        for thecls in path[1:]:
            prefix += thecls._meta.model_name + '__'
            if thecls == target_cls:
                return prefix+selector

        raise Exception('BUG: class %s not in path %s' % (target_cls, path))

    @classmethod
    def get_primary_class_from_fullnames(cls, fullnames):
        if not fullnames:
            return cls
        return cls.get_class_from_fullname(fullnames[0])[0]

    @classmethod
    def by_fullnames(cls, fullnames):
        """
        Returns a unfiltered values_list() of all given field names.
        """
        selectors = [cls.get_selector_from_fullname(f) for f in fullnames]
        primary_cls = cls.get_primary_class_from_fullnames(fullnames)
        return primary_cls.objects.values_list(*selectors)

    @classmethod
    def dom_from_query(cls, query, aliases=None):
        if not aliases:
            aliases = cls.get_aliases()
        fields = {}
        for alias in aliases:
            fields[alias] = cls.get_classname()+'.'+alias
        query_parser = QueryParser(fields, aliases)
        return query_parser.parse(query)

    @classmethod
    def q_from_query(cls, query, aliases=None):
        """
        Returns a Q-Object for the given query.
        """
        dom = cls.dom_from_query(query, aliases)
        serializer = DjangoSerializer(cls)
        return dom.serialize(serializer)

    @classmethod
    def by_query(cls, query, aliases=None):
        return cls.objects.filter(cls.q_from_query(query, aliases))

    @classmethod
    def sql_from_query(cls, query, mode='SELECT', fullnames=None, extra_model=None):
        """
        Returns an SQL statement for the given query.
        """
        dom = cls.dom_from_query(query)
        return sql_from_dom(cls, dom,
                            mode=mode,
                            fullnames=fullnames,
                            extra_model=extra_model)

    @classmethod
    def by_query_raw(cls, query, mode='SELECT', fullnames=None, extra_model=None):
        """
        Returns a PaginatedRawQuerySet for the given query.
        """
        sql, args, fields = cls.sql_from_query(query,
                                               mode=mode,
                                               fullnames=fullnames,
                                               extra_model=extra_model)
        return PaginatedRawQuerySet(cls, sql, args), fields

    @classmethod
    def sql_from_json(cls, json_string, mode='SELECT', extra_model=None):
        dom = JSONParser().parse(json_string)
        return sql_from_dom(cls, dom, extra_model=extra_model)

    @classmethod
    def by_json_raw(cls, json_string, extra_model=None):
        sql, args, fields = cls.sql_from_json(json_string,
                                              extra_model=extra_model)
        return PaginatedRawQuerySet(cls, sql, args), fields



from builtins import str
from collections import defaultdict, OrderedDict
from ..refs import get_join_for
from .serializer import Serializer
from .util import parse_date, parse_datetime

int_op_map = {
    'equals': 'equals',
    'contains': 'equals',
    'startswith': 'gte',
    'endswith': 'lte'
}

str_op_map = {
    'gt': 'startswith',
    'gte': 'startswith',
    'lt': 'endswith',
    'lte': 'endswith'
}

date_op_map = {
    'contains': 'equals',
    'startswith': 'gte',
    'endswith': 'lte'
}

operator_map = {
    'equals': "='{}'",
    'iequals': " LIKE '{}'",
    'lt': "<'{}'",
    'lte': "<='{}'",
    'gt': ">'{}'",
    'gte': ">='{}'",
    'startswith': " LIKE '{}%%'",
    'endswith': " LIKE '%%{}'",
    'contains': " LIKE '%%{}%%'",
    'regex': " RLIKE '{}'"
}

def _mkcol(tbl, name, alias):
    return tbl+'.'+name+' '+tbl+'_'+alias

def _mk_condition(db_column, operator, data):
    op = operator_map.get(operator)
    if not op:
        raise Exception('unsupported operator:' + str(operator))

    # I would prefer to use a prepared statement, but collecting arguments
    # and passing them back along the string everywhere would be awful design.
    # (Also, I didn't find any API from Django to generate a prepared statement
    # without already executing it, e.g. django.db.connection.execute())
    if isinstance(data, int):
        return db_column+op.format(data)
    elif isinstance(data, bytes):
        data = data.decode('utf-8')
    return db_column+op.format(data)

class SQLSerializer(Serializer):
    def __init__(self, model, mode='SELECT', fullnames=None, extra_model=None):
        modes = 'SELECT', 'WHERE'
        if mode not in modes:
            raise AttributeError('invalid mode: {}. Must be one of {}'.format(mode, modes))
        Serializer.__init__(self)
        self.model = model
        self.mode = mode
        self.fullnames = fullnames
        self.extra_model = extra_model

    def _create_db_column_list(self, dom):
        fullnames = self.fullnames if self.fullnames else dom.get_term_names()
        result = []
        for fullname in fullnames:
            model, alias = self.model.get_class_from_fullname(fullname)
            selector = model.get_selector_from_alias(alias)
            target_model, field = model.get_field_from_selector(selector)
            result.append((target_model, target_model._meta.db_table, field.column))
        return result

    def _create_select(self, fields):
        # Create the "SELECT DISTINCT table1.col1, table2.col2, ..."
        # part of the SQL.
        col_numbers = defaultdict(int)
        fullfields = []
        for field in fields:
            table, column = field[1:3]
            key = "%s.%s" % (table, column)
            col_number = col_numbers[key]
            col_numbers[key] += 1
            if len(field) == 3:
                field = (field[0], table, column, column if col_number == 0 else "%s__%d" % (column, col_number))
            fullfields.append(field)

        select = 'SELECT DISTINCT '+_mkcol(fullfields[0][1], fullfields[0][2], fullfields[0][3])
        for target_model, table, column, alias in fullfields[1:]:
            select += ', '+_mkcol(table, column, alias)

        # Find the best way to join the tables.
        target_models = [r[0] for r in fullfields]
        if self.extra_model:
            target_models.append(self.extra_model)
        vector = self.model.get_object_vector_for(target_models)
        join_path = get_join_for(vector)

        # Create the "table1 LEFT JOIN table2 ON table1.col1=table2.col1"
        # part of the SQL.
        select += ' FROM '+join_path[0][0]
        for table, left, right in join_path[1:]:
            select += ' LEFT JOIN {} ON {}={}'.format(table,
                                                      table+'.'+left,
                                                      right)
        return select

    def logical_root_group(self, root_group, terms):
        fields = self._create_db_column_list(root_group)

        # Create the SELECT part of the query.
        if self.mode == 'SELECT':
            select = self._create_select(fields)+' WHERE '
        else:
            select = ''

        where = (' AND '.join(terms) if terms else '1')
        if where.startswith('(') and where.endswith(')'):
            select += where
        else:
            select += '('+where+')'
        return select, []

    def logical_group(self, terms):
        terms = [t for t in terms if t]
        if not terms:
            return ''
        return ' AND '.join(terms)

    def logical_and(self, terms):
        terms = [t for t in terms if t]
        if not terms:
            return '()'
        return '(' + self.logical_group(terms) + ')'

    def logical_or(self, terms):
        terms = [t for t in terms if t]
        if not terms:
            return ''
        return '(' + ' OR '.join(terms) + ')'

    def logical_not(self, terms):
        if not terms:
            return ''
        if len(terms) == 1:
            return 'NOT(' + terms[0] + ')'
        return 'NOT ' + self.logical_and(terms)

    def boolean_term(self, db_column, operator, data):
        value = 'TRUE' if data.lower() == 'true' else 'FALSE'
        return _mk_condition(db_column, operator, value)

    def int_term(self, db_column, operator, data):
        try:
            value = int(data)
        except ValueError:
            return '1'
        operator = int_op_map.get(operator, operator)
        return _mk_condition(db_column, operator, value)

    def str_term(self, db_column, operator, data):
        operator = str_op_map.get(operator, operator)
        return _mk_condition(db_column, operator, data)

    def lcstr_term(self, db_column, operator, data):
        operator = str_op_map.get(operator, operator)
        if operator == 'equals':
            operator = 'iequals'
        return _mk_condition(db_column, operator, data.lower())

    def date_datetime_common(self, db_column, operator, thedatetime):
        if not thedatetime:
            return ''
        operator = date_op_map.get(operator, operator)
        return _mk_condition(db_column, operator, thedatetime.isoformat())

    def date_term(self, db_column, operator, data):
        thedate = parse_date(data)
        return self.date_datetime_common(db_column, operator, thedate)

    def datetime_term(self, db_column, operator, data):
        thedatetime = parse_datetime(data)
        return self.date_datetime_common(db_column, operator, thedatetime)

    def term(self, term_name, operator, data):
        if operator == 'any':
            return '1'

        model, alias = self.model.get_class_from_fullname(term_name)
        selector = model.get_selector_from_alias(alias)
        target_model, field = model.get_field_from_selector(selector)
        db_column = target_model._meta.db_table + '.' + field.column
        handler = model.get_field_handler_from_alias(alias)

        type_map = {'BOOL': self.boolean_term,
                    'INT': self.int_term,
                    'STR': self.str_term,
                    'LCSTR': self.lcstr_term,
                    'DATE': self.date_term,
                    'DATETIME': self.datetime_term}

        func = type_map.get(handler.db_type)
        if not func:
            raise TypeError('unsupported field type: '+repr(field_type))
        return func(db_column, operator, handler.prepare(data))



from copy import copy
from itertools import chain
from django.db import models
from django.db.models.fields.related import ManyToOneRel, ManyToManyRel


def _get_joining_columns(field, reverse=False):
    """Compatibility shim for Django <5.0 through 6.0+.

    Django 5.0 deprecated get_joining_columns() in favour of
    get_joining_fields() which returns field objects instead of column
    name strings.  Django 6.0 removed the old method entirely.
    """
    if hasattr(field, "get_joining_fields"):
        pairs = field.get_joining_fields(reverse_join=reverse)
        return tuple((lhs.column, rhs.column) for lhs, rhs in pairs)
    if reverse:
        return field.get_reverse_joining_columns()
    return field.get_joining_columns()

def get_subclasses(cls):
    """
    Recursively finds all subclasses of the current class.
    Like Python's __class__.__subclasses__(), but recursive.
    Returns a list containing all subclasses.

    @type cls: object
    @param cls: A Python class.
    @rtype: list(object)
    @return: A list containing all subclasses.
    """
    result = set()
    path = [cls]
    while path:
        parent = path.pop()
        for child in parent.__subclasses__():
            if not '.' in str(child):
                # In a multi inheritance scenario, __subclasses__()
                # also returns interim-classes that don't have all the
                # methods. With this hack, we skip them.
                continue
            if child not in result:
                result.add(child)
                path.append(child)
    return result

def child_classes(cls):
    """
    Returns all models that have a foreign key pointing to cls.
    """
    children = []
    for field in cls._meta.get_fields():
        if isinstance(field, (ManyToOneRel, ManyToManyRel)):
            children.append(field.related_model)
    return children

def parent_classes(cls):
    """
    Returns all models that are referenced by a foreign key of the
    given class.
    """
    parents = []
    for field in cls._meta.get_fields():
        if isinstance(field, (models.ForeignKey, models.ManyToManyField)):
            parents.append(field.remote_field.model)
    return parents

def get_field_to(cls, target_cls):
    for field in cls._meta.get_fields():
        if not isinstance(field, (models.ForeignKey, models.ManyToManyField)):
            continue
        if field.remote_field.model is target_cls:
            return field
    return None

def get_object_vector_to(cls, search_cls, subtype, avoid=None):
    """
    Returns a list of all possible paths to the given class.
    Only searches classes that are subtype of the given class.
    """
    # Does the name point to this class? Then we are done.
    if search_cls == cls:
        return [(cls,)]

    # Avoid endless recursion.
    if avoid is None:
        avoid = set()
    else:
        avoid = copy(avoid)
    avoid.add(cls)

    # So the name does not point to this class. Delegate the request to
    # each of our connected classes, collecting all possible paths.
    path_list = []
    for thecls in chain(child_classes(cls), parent_classes(cls)):
        if thecls in avoid:
            continue
        if subtype in thecls.__mro__:
            child_path_list = get_object_vector_to(thecls, search_cls, subtype, copy(avoid))
        elif thecls == search_cls:
            child_path_list = [(thecls,)]
        else:
            continue
        for path in child_path_list:
            path_list.append((cls,)+path)
    path_list.sort(key=len)
    return path_list

def yield_all_vectors(search_cls_list, subtype):
    """ 
    Yields all possible vectors between all given classes.
    """
    for target_cls in search_cls_list[:]:
        for thecls in search_cls_list:
            vectors = get_object_vector_to(thecls, target_cls, subtype)
            for vector in vectors:
                yield vector

def yield_matching_vectors(vectors, search_cls_list):
    """
    Yields all possible vectors that connect all of the given classes.
    The result is sorted by the position of the primary class, and
    the vector length.
    """
    for vector in vectors:
        for target_cls in search_cls_list:
            if target_cls not in vector:
                break
        else:
            yield vector

def sort_vectors_by_primary_cls(vectors, primary_cls):
    """
    Sort the vectors by the position of the primary class, and
    the vector length.
    """
    def sort_by_length_and_pos_of_primary_cls(vector):
        try:
            pos = vector.index(primary_cls)
        except ValueError:
            pos = 0
        return float('{}.{}'.format(pos, len(vector)))
    return sorted(vectors, key=sort_by_length_and_pos_of_primary_cls)

def get_object_vector_for(cls, search_cls_list, subtype, avoid=None):
    """
    Like get_object_vector_to(), but returns a single vector that reaches
    all of the given classes, if it exists.
    Only searches classes that are subtype of the given class.
    """
    vectors = list(yield_all_vectors(search_cls_list, subtype))
    matching = list(yield_matching_vectors(vectors, search_cls_list))
    if not matching:
        return None # No vector contains all classes

    # Prefer the path where the classes appear in the same order as in
    # search_cls_list.
    primary_cls = search_cls_list[0]
    matching = sort_vectors_by_primary_cls(matching, primary_cls)
    for vector in matching:
        # Remove extra classes that are not explicitly requested.
        clean_vector = [c for c in vector if c in search_cls_list]
        if clean_vector == search_cls_list:
            return vector

    return matching[0] # No vector contains all classes in the same order.

def _class2through_columns(from_cls, through_model):
    table_name = from_cls._meta.db_table
    through_table = through_model._meta.db_table
    through_left = get_field_to(through_model, from_cls).column
    through_right = table_name+'.'+from_cls._meta.pk.column
    return through_table, through_left, through_right

def _through2class_columns(through_model, target_cls):
    # From through-table to target class.
    through_table = through_model._meta.db_table
    from_field = get_field_to(through_model, target_cls)
    through_join = _get_joining_columns(from_field, reverse=True)
    left, right = through_join[0]
    right = through_table+'.'+right
    return left, right

def _get_through_model(cls, field):
    if isinstance(field, models.fields.related.ManyToManyField):
        return getattr(cls, field.attname).through
    return None

def _get_join_data(last_cls, to_cls):
    result = []
    last_table_name = last_cls._meta.db_table

    # Two options: The current class has a reference to the other table,
    # or the other way around.
    field = get_field_to(last_cls, to_cls)
    if field:
        through_model = _get_through_model(last_cls, field)
        if through_model:
            result.append(_class2through_columns(last_cls, through_model))
            left, right = _through2class_columns(through_model, to_cls)
        else:
            left, right = _get_joining_columns(field, reverse=True)[0]
            right = last_table_name+'.'+right

    else:
        field = get_field_to(to_cls, last_cls)
        if field is None:
            raise AttributeError('JOIN for unconnected objects is not possible')

        through_model = _get_through_model(to_cls, field)
        if through_model:
            result.append(_class2through_columns(last_cls, through_model))
            left, right = _through2class_columns(through_model, to_cls)
        else:
            left, right = _get_joining_columns(field)[0]
            right = last_table_name+'.'+right

    table_name = to_cls._meta.db_table
    result.append((table_name, left, right))
    return result

def get_join_for(vector):
    """
    Given a vector as returned by get_object_vector_for(), this function
    returns a list of tuples that explain how to join the models (tables)
    together on the SQL layer. Each tuple has three elements::

        (table_name, left_key, right_key)

    In the first tuple of the list, left_key is always None.
    In the second tuple of the list, right_key is always None.
    All other tuples required both keys to join them.

    Complete example (keep in mind that the connection between Component
    and Unit is many-to-many, so there's a helper table here)::

        get_join_path_for((Device, Component, Unit))

    This returns::

        [
            ('inventory_device', None, None),
            ('inventory_component', 'device_id', 'inventory_device.metadata_id'),
            ('inventory_unit_component', 'component_id', 'inventory_component.id'),
            ('inventory_unit', 'id', 'inventory_unit_component.unit_id')
        ]

    Which means that the following SQL JOIN could be used::

        SELECT *
        FROM inventory_device
        LEFT JOIN inventory_component ON inventory_component.device_id=inventory_device.metadata_id
        LEFT JOIN inventory_unit_component ON inventory_unit_component.component_id=inventory_component.id
        LEFT JOIN inventory_unit ON inventory_unit.id=inventory_unit_component.unit_id
    """
    result = [(vector[0]._meta.db_table, None, None)]
    for pos, thecls in enumerate(vector[1:]):
        last_cls = vector[pos]
        result += _get_join_data(last_cls, thecls)
    return result



from django.db import models

class FieldHandler(object):
    """
    Abstract base type for all field handlers.

    A field handler is an object that you can use to define custom
    behavior when searching a field of a model.

    You might want to use a field handler if you are using a custom
    model field, or if your query contains information that
    requires client-side processing before being passed to
    the database.
    """
    db_type = None

    @classmethod
    def handles(cls, model, field):
        raise NotImplemented

    @classmethod
    def prepare(cls, value):
        return value

class StrFieldHandler(FieldHandler):
    db_type = 'STR'

    @classmethod
    def handles(cls, model, field):
        return isinstance(field, (models.CharField, models.TextField))

class LowerCaseStrFieldHandler(StrFieldHandler):
    db_type = 'LCSTR'

class IPAddressFieldHandler(LowerCaseStrFieldHandler):
    @classmethod
    def handles(cls, model, field):
        return isinstance(field, models.GenericIPAddressField)

class BooleanFieldHandler(FieldHandler):
    db_type = 'BOOL'

    @classmethod
    def handles(cls, model, field):
        return isinstance(field, models.BooleanField)

class IntegerFieldHandler(FieldHandler):
    db_type = 'INT'

    @classmethod
    def handles(cls, model, field):
        return isinstance(field, (models.IntegerField, models.AutoField))

class DateFieldHandler(FieldHandler):
    db_type = 'DATE'

    @classmethod
    def handles(cls, model, field):
        return isinstance(field, models.DateField)

class DateTimeFieldHandler(FieldHandler):
    db_type = 'DATETIME'

    @classmethod
    def handles(cls, model, field):
        return isinstance(field, models.DateTimeField)

type_registry = [
        LowerCaseStrFieldHandler,
        IPAddressFieldHandler,
        BooleanFieldHandler,
        IntegerFieldHandler,
        DateTimeFieldHandler,
        DateFieldHandler,
]
